#property copyright "Scriptong"
#property link "http://autograf.dp.ua"

#property indicator_chart_window                   //     

extern double PSARStep          = 0.01;
extern double PSARMax           = 0.19;
extern color  colorUpwardWave   = Blue;
extern color  colorDownwardWave = Red;
extern color  textColor         = DeepPink;
extern int    textSize          = 14;
extern string textFont          = "Tahoma";

double changeRegData[];                            //    /..
                                                   // ..  
                                                   // ..
double changeFactData[];                           //    /..
                                                   // ..  
double minMaxBuf[];                                //     ..
                                                   // ..    
double waveNumber[];                               //                                                                                                       

                                                   
#define NO_WAVE         0                          //  
#define WAVE_1          1                          //  1 -  
#define WAVE_2          2                          //  2 -  
#define WAVE_3          3                          //  3 -  
#define WAVE_4          4                          //  4 -  
#define WAVE_5          5                          //  5 -  
#define WAVE_A          6                          //  A -  
#define WAVE_B          7                          //  B -  
#define WAVE_C          8                          //  C -  

#define TREND_CHANGE_FROM_UP_TO_DOWN 1             //     ..
                                                   // ..
#define TREND_CHANGE_FROM_DOWN_TO_UP 2             //     ..
                                                   // ..
#define TREND_FOLLOW_UP              3             //   
#define TREND_FOLLOW_DOWN            4             //   

string prefix = "WOPSAR_";

//+-------------------------------------------------------------------------------------+
//| Custom indicator initialization function                                            |
//+-------------------------------------------------------------------------------------+
int init()
{
   IndicatorBuffers(4);
// - 1 - ==       ============
   string name = WindowExpertName();
// - 1 - ==   =============================================================

// - 2 - ==    ,   ==========================
   SetIndexBuffer(0, changeRegData);               //   - /..
                                                   // ..  ..
                                                   // ..
   SetIndexBuffer(1, changeFactData);              //   - /..
                                                   // ..  
   SetIndexBuffer(2, minMaxBuf);                   //   -  ..
                                                   // ..
   SetIndexBuffer(3, waveNumber);                  //   -  ..
                                                   // ..
// - 2 - ==   =============================================================

   return(0);
}
//+-------------------------------------------------------------------------------------+
//| Custom indicator deinitialization function                                          |
//+-------------------------------------------------------------------------------------+
int deinit()
{
   for (int i = ObjectsTotal()-1; i >= 0; i--)     //    
      if (StringSubstr(ObjectName(i), 0, StringLen(prefix)) == prefix)
         ObjectDelete(ObjectName(i));

   return(0);
}
//+-------------------------------------------------------------------------------------+
//|   ,                   |
//+-------------------------------------------------------------------------------------+
int GetRecalcIndex()
{
   int counted_bars = IndicatorCounted();
   if (counted_bars == 0)                          // -   - 0. 
   {                                               // ..    ..
      ArrayInitialize(changeRegData, EMPTY_VALUE); // ...  
      ArrayInitialize(changeFactData, EMPTY_VALUE);
      ArrayInitialize(minMaxBuf, EMPTY_VALUE);
      ArrayInitialize(waveNumber, 0);
      return(Bars - 2);                            //     
   }
   return(Bars - counted_bars - 1);                //    
}
//+-------------------------------------------------------------------------------------+
//|   waveNumber                                           |
//+-------------------------------------------------------------------------------------+
void FillTheBufferWithNewData(int start, int end, int newData)
{
   for (int i = start; i >= end; i--)
      waveNumber[i] = newData;
}
//+-------------------------------------------------------------------------------------+
//|                                                |
//+-------------------------------------------------------------------------------------+
int GetNumberOfWave(double latePSAR, double earlyPSAR, int index1, int index2,
                    bool isPrevWaveUpward = true)
{
// - 1 - ==     ============================================
   if ((latePSAR >= earlyPSAR && isPrevWaveUpward) ||//    ..
       (latePSAR <= earlyPSAR && !isPrevWaveUpward))// ..   
   {                  
      if (waveNumber[index1] == WAVE_1 ||          //    ..
          waveNumber[index1] == WAVE_3 ||          // ..   ,..
          waveNumber[index1] == WAVE_5 ||          // .. ,  ..
          waveNumber[index1] == WAVE_A)            // ..   (2, 4,
         return(waveNumber[index1] + 1);           // ..B)   (A)
         
      if (waveNumber[index1] == WAVE_C)            //    ,  ..
         return(WAVE_1);                           // ..  ..
                                                   // ..
                                          
      FillTheBufferWithNewData(index1, index2, WAVE_1);//     ..
      return(WAVE_2);                              // ..  ..
                                                   // ...  ..
                                                   // ..  ..
                                                   // ..   
   }
// - 1 - ==   =============================================================
   
// - 2 - ==     () =============================
   if (waveNumber[index1] == WAVE_2 ||             //    ..
       waveNumber[index1] == WAVE_4 ||             // ..  .
       waveNumber[index1] == WAVE_B)               //  ,   ..
      return(waveNumber[index1] + 1);              // .. 

   FillTheBufferWithNewData(index1, index2, NO_WAVE);//    ..
                                                   // ..  , ..
                                                   // ..    . 
                                                   //    ..
   return(WAVE_1);                                 // ..    
// - 2 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//|                                         |
//+-------------------------------------------------------------------------------------+
string GetWaveFromIndex(int wave)
{
   switch(wave)
   {
      case WAVE_1: return("I");
      case WAVE_2: return("II");
      case WAVE_3: return("III");
      case WAVE_4: return("IV");
      case WAVE_5: return("V");
      case WAVE_A: return("A");
      case WAVE_B: return("B");
      case WAVE_C: return("C");
   }
   return("UD");                                   //    
}
//+-------------------------------------------------------------------------------------+
//|                                  |
//+-------------------------------------------------------------------------------------+
void CreateWave(datetime createTime, datetime time1, double price1, datetime time2, 
                double price2, int waveNumber, color clr)
{
// - 1 - ==    =====================================================
   string name = prefix + "line_" + createTime;    //  
   if (ObjectFind(name) < 0)                       //    , ..
   {                                               // .. 
      ObjectCreate(name, OBJ_TREND, 0, time1, price1, time2, price2);
      ObjectSet(name, OBJPROP_RAY, false);
      ObjectSet(name, OBJPROP_COLOR, clr);
   }
   else                                            //   , ..
   {                                               // ..    
      ObjectMove(name, 0, time1, price1);
      ObjectMove(name, 1, time2, price2);
      ObjectSet(name, OBJPROP_COLOR, clr);
   }
// - 1 - ==   =============================================================
   
// - 2 - ==     =======================================
   name = prefix + "letter_" + createTime;         //  
   datetime letterTime = time1 + MathFloor((time2 - time1)/2);//    - 
                                                   // ..   ..
                                                   // ..   
   double letterPrice = (price1 + price2)/2;       //  Y -  ..
                                                   // ..    
   if (ObjectFind(name) < 0)                       //    , ..
   {                                               // .. 
      ObjectCreate(name, OBJ_TEXT, 0, letterTime, letterPrice);
      ObjectSetText(name, GetWaveFromIndex(waveNumber), textSize, textFont, textColor);
      return;
   }                                               //   , ..
   ObjectMove(name, 0, letterTime, letterPrice);   // ..    
   ObjectSetText(name, GetWaveFromIndex(waveNumber), textSize, textFont, textColor);
// - 2 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//|                                                      |
//+-------------------------------------------------------------------------------------+
void RefreshWave(datetime time1, datetime time2, double price2)
{
// - 1 - ==     ==========================================
   string name = prefix + "line_" + time1;
   if (ObjectFind(name) == 0)
      ObjectMove(name, 1, time2, price2);
// - 1 - ==   =============================================================

// - 2 - ==     =======================================
   datetime leftTime = ObjectGet(name, OBJPROP_TIME1);//     Y ..
   datetime letterTime = leftTime + MathFloor((time2 - leftTime)/2);// ..
   double letterPrice = (ObjectGet(name, OBJPROP_PRICE1) + price2)/2;
   name = prefix + "letter_" + time1;                 //   
   if (ObjectFind(name) == 0)                         //  
   {
      ObjectSetText(name, ObjectDescription(name), textSize, textFont, textColor);
      ObjectMove(name, 0, letterTime, letterPrice);
   }
// - 2 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//|                                                     |
//+-------------------------------------------------------------------------------------+
color GetWaveColor(int waveNumber, bool isUpwardWave = true)
{
// - 1 - ==    =======================================================
   if (isUpwardWave)                                  //    ..
   {                                                  // ..  ..
      if (waveNumber%2 == 0)                          // ..  ..
         return(colorDownwardWave);                   // ..    ..
      return(colorUpwardWave);                        // ..  
   }                                                  // ..
// - 1 - ==   =============================================================
   
// - 2 - ==    =======================================================
   if (waveNumber%2 == 0)                             //    ..
      return(colorUpwardWave);                        // ..   
   return(colorDownwardWave);                         // ..    ..
                                                      // ..   
                                                      // ..  
// - 2 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//|                                               |
//+-------------------------------------------------------------------------------------+
void DrawOneWave(int lateIndex, int earlyIndex, bool isUpwardWave = true)
{  
// - 1 - ==      ,     2- =====
   if (waveNumber[lateIndex] < WAVE_2)
      return;
// - 1 - ==   =============================================================

// - 2 - ==    ====================================================
   color clr = GetWaveColor(waveNumber[lateIndex], isUpwardWave);//  
   if (waveNumber[lateIndex] == WAVE_2)            //    ..
      CreateWave(changeRegData[lateIndex+1],       // ..,    
                 changeFactData[earlyIndex+1],     // ..  
                 minMaxBuf[earlyIndex+1], 
                 changeFactData[lateIndex+1], 
                 minMaxBuf[lateIndex+1],
                 WAVE_1, 
                 clr);
// - 2 - ==   =============================================================
   
// - 3 - ==    ====================================================
   CreateWave(changeRegData[lateIndex],            //    
              changeFactData[lateIndex+1],         //   ( ) 
              minMaxBuf[lateIndex+1],              //   
              changeFactData[lateIndex],           //   
              minMaxBuf[lateIndex],                //   
              waveNumber[lateIndex],               //  
              clr);                                //  
// - 3 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//|     :                    |
//+-------------------------------------------------------------------------------------+
int GetBarSituation(int index)
{
// - 1 - ==     ==========================================
   double psar1 = iSAR(NULL, 0, PSARStep, PSARMax, index);
   double psar2 = iSAR(NULL, 0, PSARStep, PSARMax, index+1);
// - 1 - ==   =============================================================

   //      :
   //  1 -      
   //  2 -      
   //  3 -   
   //  4 -   
   
   //          ..
   // ...  ,      . 
   //          ..
   // .. , ..   ,    ..
   // ..

// - 2 - ==       ================================
   if (Close[index] > psar1)
   {                                               //   :
      if (Close[index+1] < psar2)                  //  1 -   
         return(TREND_CHANGE_FROM_DOWN_TO_UP);     //  ..  -  
      if (Close[index+1] > psar2)                  //  2 -   
         return(TREND_FOLLOW_UP);                  //  ..  - 
      for (int i = index+2; i < Bars; i++)         //  3 -    ..
      {                                            //  .. -  ,..
         double psarN = iSAR(NULL, 0, PSARStep, PSARMax, i);// ..  ..
         if (Close[i] != psarN)                    //  ..  .
            break;
      }
      if (i == Bars || Close[i] > psarN)           //   , ,..
         return(TREND_FOLLOW_UP);                  // .. ..
      return(TREND_CHANGE_FROM_DOWN_TO_UP);        // .. 
   }
// - 2 - ==   =============================================================

// - 3 - ==       ================================
   if (Close[index] < psar1)
   {                                               //    :
      if (Close[index+1] > psar2)                  //  1 -   
         return(TREND_CHANGE_FROM_UP_TO_DOWN);     //  ..  -  
      if (Close[index+1] < psar2)                  //  2 -   
         return(TREND_FOLLOW_DOWN);                //  ..  - 
      for (i = index+2; i < Bars; i++)             //  3 -    ..
      {                                            //  .. -  ,..
         psarN = iSAR(NULL, 0, PSARStep, PSARMax, i);// ..  ..
         if (Close[i] != psarN)                    //  ..  .
            break;
      }
      if (i == Bars || Close[i] < psarN)           //   , ,..
         return(TREND_FOLLOW_DOWN);                // .. ..
      return(TREND_CHANGE_FROM_UP_TO_DOWN);        // .. 
   }
// - 3 - ==   =============================================================

// - 4 - ==       ===============================
   //          ..
   // ..
   for (i = index+1; i < Bars; i++)                //  ,   ..
   {                                               // ..  
      psarN = iSAR(NULL, 0, PSARStep, PSARMax, i);
      if (Close[i] != psarN)
         break;
   }
   if (i == Bars || Close[i] < psarN)              //   , ,..
      return(TREND_FOLLOW_DOWN);                   // ..  ..
   return(TREND_FOLLOW_UP);                        // .. 
// - 4 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//|                                                |
//+-------------------------------------------------------------------------------------+
void ChangeTrend(int index)
{
// - 1 - ==   =======================================================
   int barSituation = GetBarSituation(index);     //        
   if (changeRegData[index+1] == EMPTY_VALUE)     //     , ..
   {                                              // ..   ..
      changeRegData[index+1] = Time[index+1];     // ..
      changeFactData[index+1] = Time[index+1];
      minMaxBuf[index+1] = High[index+1];
   }
   changeRegData[index] = changeRegData[index+1]; //   ,  .. 
   changeFactData[index] = changeFactData[index+1];// ..    
   minMaxBuf[index] = minMaxBuf[index+1];
   waveNumber[index] = waveNumber[index+1];
// - 1 - ==   =============================================================

// - 2 - ==    ===============================================
   if (barSituation < TREND_FOLLOW_UP)             //     ..
   {                                               // ..   ..
                                                   // .. 
   // - 2.1 - ==       ===============================
      changeRegData[index] = Time[index];          //   
      changeFactData[index] = Time[index];         //    ..
                                                   // .. (, .. ..
                                                   // .. )
      double psar2 = iSAR(NULL, 0, PSARStep,       //  PSAR   
                          PSARMax, index+1);       
      int prevChangeIndex = iBarShift(NULL, 0,     //   ..
                                     changeRegData[index+1]);// ..  
      double psarN = iSAR(NULL, 0, PSARStep,       //  PSAR   ..
                          PSARMax, prevChangeIndex+1);// ..  
   // - 2.1 - ==   ========================================================

   // - 2.2 - ==       ===============================
      if (barSituation == TREND_CHANGE_FROM_DOWN_TO_UP)//   - 
      {
         minMaxBuf[index] = High[index];           //   
         waveNumber[index] = GetNumberOfWave(psar2,//    
                                             psarN, prevChangeIndex, index+1, false);
         DrawOneWave(index, prevChangeIndex);      //              
      }
   // - 2.2 - ==   ========================================================

   // - 2.3 - ==       ===============================
      if (barSituation == TREND_CHANGE_FROM_UP_TO_DOWN)//   - 
      {                                            
         minMaxBuf[index] = Low[index];            //   
         waveNumber[index] = GetNumberOfWave(psar2,//    
                                             psarN, prevChangeIndex, index+1);
         DrawOneWave(index, prevChangeIndex, false);//              
      }
   // - 2.3 - ==   ========================================================
      return;
   }
// - 2 - ==   =============================================================
   
// - 3 - ==    ============================================
   if (barSituation == TREND_FOLLOW_DOWN)
      if (minMaxBuf[index+1] == EMPTY_VALUE ||     //     
          Low[index] <= minMaxBuf[index+1])        // ..  
      {
         minMaxBuf[index] = Low[index];            //  
         changeFactData[index] = Time[index];      //   ..
                                                   // .. 
         RefreshWave(changeRegData[index],         //   
                     Time[index], Low[index]);
         return;
      }
// - 3 - ==   =============================================================
   
// - 4 - ==    ============================================
   if (barSituation == TREND_FOLLOW_UP)
      if (minMaxBuf[index+1] == EMPTY_VALUE ||     //    ..
          High[index] >= minMaxBuf[index+1])       // ..   
      {
         minMaxBuf[index] = High[index];           //  
         changeFactData[index] = Time[index];      //   ..
                                                   // .. 
         RefreshWave(changeRegData[index],         //   
                     Time[index], High[index]);
      }
// - 4 - ==   =============================================================
}
//+-------------------------------------------------------------------------------------+
//| Custom indicator iteration function                                                 |
//+-------------------------------------------------------------------------------------+
int start()
{
   int limit = GetRecalcIndex();                   //    
  
   for (int i = limit; i > 0; i--)
      ChangeTrend(i);
   
   WindowRedraw();
   return(0);
}